/*
	iCook iBook Overclocker
	Copyright (c) 2004 Patrick Swieskowski <pat@swieskowski.net>
	This could potentially damage your computer.  Use at your own risk.
	
	iCook is a kernel extension for OS X that lets users control the speed of their IBM
	750FX processor.  All iBooks with 16MB or more of video RAM use the 750FX and should
	be compatible with this software (Note: This does not include the iBook G4).  It
	creates a new sysctl "hw.cpuspeed" that the current cpu speed can be read from and
	new cpu speeds can be written to.  It supports speeds from 300-1200 MHz in 50 MHz
	increments.

	For this to work iCook must disable the kernel's own internal control over the cpu
	speed.  This is done by patching the kernel image in memory.  No permanent changes are
	made and a reboot will bring the system back to normal.
	
	IMPORTANT: A side effect of this is that the Processor Performance setting in OS X's
	Energy Saver preferences will no longer function properly once iCook is loaded.  It is 
	recommended that this be set to Highest before using iCook.  This will configure other
	system components to work with a high-speed processor and allow you to get the best
	performance from your system.  The Reduced setting works with iCook, but the system will
	start to become unstable at a lower speed.  Using the Automatic setting with iCook is
	just a bad idea, don't do this.
	
	Example Usage:
		$ sudo kextload -b net.swieskowski.iCook   
		kextload: /System/Library/Extensions/iCook.kext loaded successfully
		$ sysctl hw.cpuspeed
		hw.cpuspeed: 700
		$ sudo sysctl -w hw.cpuspeed=800
		hw.cpuspeed: 700 -> 800
		
	For more information see http://www.swieskowski.net/icook/
*/
 
 /*
	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <mach/mach_types.h>
#include <sys/systm.h>
#include <sys/types.h>
#include <sys/sysctl.h>
#include <ppc/spl.h>
#include <ppc/proc_reg.h>

/* processor identification constants */
#define PVR_MASK  0xffff0000
#define PVR_750FX 0x70000000

/* some constants for our kernel patching routine */
#define NOP_INSTR 0x60000000		/* nop instruction */
#define BAD_INSTR 0x7c11fba6		/* mtspr hid1, r? template */
#define BAD_MASK  0xfc1fffff		/* mask for comparison with BAD_INSTR */

/* fields in HID1, see IBM's 750FX user's manual */
#define HID1_PSTAT_BIT		7		/* odd, this doesn't behave as defined, we'll just use PS instead */
#define HID1_PI0_BIT		14
#define HID1_PS_BIT			15
#define HID1_PR1_BIT		29

/* how long it takes a PLL to lock, in microseconds */
/* datasheet says 100, but we'll double it to be safe */
#define PLL_LOCKTIME		200

extern void delay(uint64_t microseconds);

static int iCook_cpuspeed; /* variable for the sysctl */

static int iCook_getspeed(void)
{
    u_int32_t hid1;
    
    asm ("mfspr %0,hid1;" : "=r"(hid1));
    
    if (hid1 & MASK(HID1_PS)) {
            /* using PLL1 */
            return 50*((hid1 >> (31-28)) & 0x1f);
    } else {
            /* using PLL0 */
            if (hid1 & MASK(HID1_PI0)) {
                    /* using internal fields */
                    return 50*(hid1 >> (31-20) & 0x1f);
            } else {
                    /* using external configuration */
                    return 50*(hid1 >> (31-4) & 0x1f);
            }
    } 
}

static void iCook_setknowngood(void)
{
    u_int32_t hid1;

    asm ("mfspr %0,hid1;" : "=r"(hid1));
	
	if ( (hid1 & MASK(HID1_PS)) == 0 ) {
		/* we're running off of PLL0, that's fine */
		return;
	}
	
	/* okay, we'll have to switch to PLL0 */
	
	/* first make sure we're using the external configuration bits since these should always be sane */
	if ( (hid1 & MASK(HID1_PI0)) ) { 
		hid1 &= ~(MASK(HID1_PI0));
		asm ("mtspr hid1,%0;" : : "r"(hid1));
		delay(PLL_LOCKTIME);  /* let PLL0 lock */
	}

	/* now select PLL0 for source */
	hid1 &= ~(MASK(HID1_PS));
	asm ("mtspr hid1,%0;" : : "r"(hid1));

	return;
}

static int iCook_setspeed(int speed)
{
    u_int32_t hid1;
	int spl;
    
	if (speed < 300 || speed > 1200 || speed % 50 != 0)
		return EINVAL;
	
	spl = splhigh(); /* turn off interrupts, CPU will usually crash if someone else is playing with HID1 */
	
	iCook_setknowngood(); /* get us running off of a sane PLL0 */

    asm ("mfspr %0,hid1;" : "=r"(hid1));
    
    hid1 &= 0xfffeff01; /* mask off PLL1 config and range, also make certain that we aren't using PLL1 */
	
	if (speed < 600)
		hid1 |= MASK(HID1_PR1); /* set range bit if needed */
	
    speed /= 50;
    speed <<= 3;
    hid1 |= (speed & 0x000000f8);

    /* save settings and wait for PLL1 to lock */
	asm ("mtspr hid1,%0;" : : "r"(hid1));
	delay(PLL_LOCKTIME);
    
    hid1 |= MASK(HID1_PS);
    /* switch to PLL1 */
    asm ("mtspr hid1,%0;" : : "r"(hid1));

	splx(spl); /* turn interrupts back on */
	
    return 0;
}

static int iCook_sysctl SYSCTL_HANDLER_ARGS
{
    int speed, ret=0;
    
    if (req->newptr == NULL) {
        /* reqd request */
        speed = iCook_cpuspeed = iCook_getspeed();
        SYSCTL_OUT(req, &speed, sizeof(speed));
    } else {
        /* write request */
        SYSCTL_IN(req, &speed, sizeof(speed));
        ret = iCook_setspeed(speed);
    }
   
    return ret;
}

static void iCook_patchzone(vm_offset_t start, vm_offset_t end)
{
	int count = 0;
	u_int32_t msr, newmsr;
	
	asm("mfmsr %0;" : "=r"(msr));
	newmsr = msr & ~(MASK(MSR_EE) | MASK(MSR_ME) | MASK(MSR_DR));
	/* this will ensure that we don't get interrupted and that we can do whatever we want to memory */
	/* however it stops us from accessing our own data, so we'll have to put everything in registers */
	
	asm volatile (
					/* load all data into registers - when we switch off the MMU it will be inaccessible */
					"lwz r2,%0;\n"					/* r2 = current address */
					"lwz r3,%1;\n"					/* r3 = top address */
					"addis r4,0,hi16(%2);\n"
					"ori r4,r4,lo16(%2);\n"			/* r4 = BAD_MASK */
					"addis r5,0,hi16(%3);\n"
					"ori r5,r5,lo16(%3);\n"			/* r5 = BAD_INSTR */
					"addis r6,0,hi16(%4);\n"
					"ori r6,r6,lo16(%4);\n"			/* r6 = NOP_INSTR */
					"lwz r7,%5;\n"					/* r7 = the new MSR with EE, ME, and DR off */
					"lwz r8,%6;\n"					/* r8 = the original MSR */
					"lwz r9,%7;\n"					/* r9 = our counter */
					
					"mtmsr r7;\n"					/* set the new MSR */
					"isync;\n"						/* mtmsr must be followed by a synchronizing instruction */
					
					"0:\n"							/* label 0 - top of loop */
					"lwz r10,0(r2);\n"				/* r10 = instruction at current address */
					"and r10,r10,r4;\n"				/* mask it */
					"cmplw r10,r5;\n"				/* compare with BAD_INSTR */
					"bne 1f;\n"						/* if not a match, skip ahead to increment current address */
					
						"addi r9,r9,1;\n"			/* otherwise, increment our count */
						"stw r6,0(r2);\n"			/* overwrite BAD_INSTR with NOP_INSTR */
						"icbi 0,r2;\n"				/* and clear the instruction cache block */
						
					"1:\n"							/* label 1 - continue loop */
					"addi r2,r2,4;\n"				/* increment current address to the next instruction */
					"cmplw r2,r3;\n"				/* compare with top address */
					"blt 0b;\n"						/* if we're not there yet, do it all again */
					
					"mtmsr r8;\n"					/* restore the old MSR */
					"isync;\n"						/* mtmsr must be followed by a synchronizing instruction */
										
					"stw r9,%7;\n"					/* save the counter for posterity */
					
					: /* no output operands */
					: "m"(start), "m"(end), "i"(BAD_MASK), "i"(BAD_INSTR), "i"(NOP_INSTR),
					  "m"(newmsr), "m"(msr), "m"(count)
					: "r2", "r3", "r4", "r5", "r6", "r7", "r8", "r9", "r10", "cc", "memory"
	);			
	
	printf("[iCook] patched %d instructions in zone 0x%08x-0x%08x\n", count, start, end);
	
	return;
}

SYSCTL_PROC(_hw, OID_AUTO, cpuspeed, CTLTYPE_INT | CTLFLAG_RW, &iCook_cpuspeed, 0, &iCook_sysctl, "I", "iCook iBook Overclocker");

kern_return_t iCook_start (kmod_info_t * ki, void * d) {
	u_int32_t pvr;
	extern vm_offset_t etext; /* this is the end of the kernel's __TEXT segment and is set up in osfmk/ppc/ppc_vm_init.c */

	/* verify that the cpu is a 750FX */
	asm ("mfspr %0,pvr;" : "=r"(pvr));
	if ((pvr & PVR_MASK) != PVR_750FX) {
		printf("[iCook] cpu is not a 750FX, giving up\n");
		return KERN_FAILURE;
	}
	
	iCook_patchzone(0, etext); /* this patches the interrupt handlers and the whole __TEXT segment */
    sysctl_register_oid(&sysctl__hw_cpuspeed);
    return KERN_SUCCESS;
}

kern_return_t iCook_stop (kmod_info_t * ki, void * d) {
    sysctl_unregister_oid(&sysctl__hw_cpuspeed);
    return KERN_SUCCESS;
}